==> aim-transport-0.9.17.tar.gz: aim-transport-0.9.17/src/buddies.c <==
#include "aimtrans.h"

struct buddy_clean_data
{
    ati ti;
    xmlnode node;
};

void _at_buddies_unsubscribe(xht ht, const char *key, void *data, void * arg)
{
    xmlnode pres;
    at_session s;
    ati ti;

    s = (at_session)arg;
    ti = s->ti;

    pres = jutil_presnew(JPACKET__UNSUBSCRIBE, jid_full(s->cur), "Transport Removal");
    xmlnode_put_attrib(pres, "from", key);
    deliver(dpacket_new(pres), ti->i);
}

result at_buddy_pending_clean(void *arg)
{
    struct buddy_clean_data *bcd;
    ati ti;
    char *user;
    xmlnode node;

    bcd = (struct buddy_clean_data *)arg;
    node = bcd->node;
    ti = bcd->ti;
    user = xmlnode_get_attrib(node, "user");

    log_debug(ZONE, "[AT] Cleaning pending for %s: %s", user, xmlnode2str(node));
    pth_mutex_acquire(&ti->buddies_mutex, 0, NULL);

    xhash_zap(ti->pending__buddies, user);

    xmlnode_free(node);
    pth_mutex_release(&ti->buddies_mutex);

    return r_UNREG;
}

void at_buddy_addtolist(at_session s, spool sp, xmlnode x)
{
    at_buddy new;
    at_buddy old;
    xmlnode item;

    for(item = xmlnode_get_firstchild(x); item != NULL; 
                    item = xmlnode_get_nextsibling(item))
    {
        char *sn;

        sn = at_normalize(xmlnode_get_attrib(item, "name"));
        old = xhash_get(s->buddies, sn);
        if(old != NULL)
        {
            log_debug(ZONE, "[AT] We already have %s in our list", sn);
            continue;
        }

 		log_debug(ZONE, "[AIM] Adding a buddy %s from xdb", sn);

        spooler(sp, sn, "&", sp);
        new = pmalloco(s->p, sizeof(_at_buddy));
        new->full = jid_new(s->p, s->ti->i->id);
        jid_set(new->full, sn, JID_USER);
        new->last = xmlnode_new_tag_pool(s->p,"query");
        new->is_away = -1;
        xmlnode_put_attrib(new->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(new->last,"stamp",jutil_timestamp()); 

        xhash_put(s->buddies, new->full->user, new);
    }
}

char *at_buddy_buildlist(at_session s, jid from)
{
    char *list;
	char *blist;
	spool sp;
    pool p, bp;
    xmlnode x;
    xmlnode msg;

    p = pool_new();
	sp = spool_new(p);

    x = at_xdb_get(s->ti, from, AT_NS_ROSTER);
    if(x != NULL)
    {
        at_buddy_addtolist(s, sp, x);
    }

    x = xhash_get(s->ti->pending__buddies, jid_full(from));
    if(x == NULL)
    {
        return NULL;
    } else {
        at_buddy_addtolist(s, sp, x);
    }
    
    list = spool_print(sp);
    if(list)
    {
        blist = strdup(list);
    } else {
        blist = NULL;
    }
    log_debug(ZONE, "[AT] Buddylist generation complete");

    pool_free(p);

    return blist;
}

int at_buddy_subscribe(ati ti, jpacket jp)
{
    xmlnode dup,dup2;
    at_session s=at_session_find_by_jid(ti, jp->from);

    /* check for a valid session */
    if(s==NULL)
    {
        xmlnode err,error;
        /* if no session, and subscribing to a buddy, not aim-t */
        err=xmlnode_new_tag("message");
        xmlnode_put_attrib(err,"type","error");
        xmlnode_put_attrib(err,"from",ti->i->id);
        xmlnode_put_attrib(err,"to",jid_full(jp->from));
        error=xmlnode_insert_tag(err,"error");
        xmlnode_insert_cdata(error,"Cannot Subscribe to a AIM Buddy without a registration",-1);
        xmlnode_put_attrib(error,"code","407");
        deliver(dpacket_new(err), ti->i);

        return 0;
    }
    /* XXX
	 * Create the duplicate for the subscribe notice
	 * Construct a subscribed notification
	 * 
	 * I had to do these in this odd order due to memory concerns
	 * Dunno why I didn't just make two dups.
	 *
	 */
	dup = xmlnode_dup(jp->x);
	dup2 = xmlnode_dup(jp->x);

	xmlnode_put_attrib(dup, "to", jid_full(jp->from));
	xmlnode_put_attrib(dup, "from", jid_full(jp->to));
	xmlnode_put_attrib(dup, "type", "subscribed");
	log_debug(ZONE, "[AIM] Sending subscribed notice\n");
    deliver(dpacket_new(dup), ti->i);

	/* Construct the subscribe back */
	xmlnode_put_attrib(dup2, "type", "subscribe");
    xmlnode_put_attrib(dup2, "to", jid_full(jp->from));
    xmlnode_put_attrib(dup2, "from", jid_full(jp->to));
	log_debug(ZONE, "[AIM] Asking for a subscribe\n");
    deliver(dpacket_new(dup2), ti->i);

	return 0;
}

int at_buddy_add(ati ti, jpacket jp)
{
	at_session s;
	at_buddy buddy;
    pool p;
	int newbud=0;
    struct buddy_clean_data *bcd;
	
    log_debug(ZONE, "[AIM] Going to add a buddy\n");
	s = at_session_find_by_jid(ti, jp->from);
	if(s==NULL || s->loggedin == 0) 
    {
        xmlnode cur, item;
        
        pth_mutex_acquire(&ti->buddies_mutex, 0, NULL);
        cur = xhash_get(ti->pending__buddies, jid_full(jp->from));
        if(cur == NULL)
        {
            cur = xmlnode_new_tag("buddies");
            xmlnode_put_attrib(cur, "user", jid_full(jp->from));
            bcd = pmalloco(xmlnode_pool(cur), sizeof(struct buddy_clean_data));
            bcd->node = cur;
            bcd->ti = ti;
            register_beat(30, at_buddy_pending_clean, bcd);
        }
        item = xmlnode_insert_tag(cur, "item");
        xmlnode_put_attrib(item, "name", jp->to->user);
        xmlnode_free(jp->x);
        log_debug(ZONE, "[AT] Pending for %s: %s", xmlnode_get_attrib(cur, "user"), xmlnode2str(cur));
        xhash_put(ti->pending__buddies, xmlnode_get_attrib(cur, "user"), cur);
        pth_mutex_release(&ti->buddies_mutex);
        return 1;
	}
    
    if(xhash_get(s->buddies, jp->to->user) == NULL)
    {
		log_debug(ZONE, "[AIM] Adding buddy %s to %s\n", jp->to->user, jid_full(jp->from));
        buddy = pmalloco(s->p, sizeof(_at_buddy));
        buddy->full=jid_new(s->p, jid_full(jp->to));
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        buddy->is_away = -1;
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());
        xhash_put(s->buddies, buddy->full->user, buddy);
        at_buddy_subscribe(ti, jp);
    }
    aim_add_buddy(s->ass, aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS),
                  jp->to->user);
    
    xmlnode_free(jp->x);
    return 1;
}

int at_parse_oncoming(struct aim_session_t *ass, 
                      struct command_rx_struct *command, ...)
{

    xmlnode away;
    xmlnode status;
	xmlnode x;
	at_session s;
    at_buddy buddy;
    ati ti;
    jpacket jp;
    char *msg;
    struct aim_userinfo_s *userinfo;
    int was_away;

    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    log_debug(ZONE, "Oncoming buddy %s", userinfo->sn);

	s = (at_session)ass->aux_data;
    ti = s->ti;

    buddy = xhash_get(s->buddies, userinfo->sn);
    if(buddy == NULL)
    {
        jid jtmp;

        buddy = pmalloco(s->p, (sizeof(_at_buddy)));
        buddy->full = jid_new(s->p, ti->i->id);
        jid_set(buddy->full, userinfo->sn, JID_USER);
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        buddy->is_away = -1;
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());

        xhash_put(s->buddies, buddy->full->user, buddy);
    }
    if(buddy->login_time == 0)
    {
        buddy->login_time = userinfo->onlinesince;
    }
    buddy->idle_time = userinfo->idletime;

    was_away = buddy->is_away;
    if(userinfo->flags&AIM_FLAG_AWAY) 
        buddy->is_away = 1;
    else
        buddy->is_away = 0;

    if(buddy->is_away == 0 && (buddy->is_away != was_away))
    {
        x = xmlnode_new_tag("presence");
	    xmlnode_put_attrib(x, "to", jid_full(s->cur));
    	xmlnode_put_attrib(x, "from", ti->i->id);
	    jp = jpacket_new(x);
    	jid_set(jp->from, at_normalize(userinfo->sn), JID_USER);
	    xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));
        status = xmlnode_insert_tag(x, "status");
        xmlnode_insert_cdata(status, "Online", -1);
        deliver(dpacket_new(jp->x), ti->i);
    } 
    else if(buddy->is_away == 1 && (buddy->is_away != was_away))
    {
        log_debug(ZONE, "[AT] Requesting Away message for %s", userinfo->sn);
        aim_getinfo(ass, command->conn, userinfo->sn, AIM_GETINFO_AWAYMESSAGE);
    }

	return 1;
}

int at_parse_offgoing(struct aim_session_t *ass, 
                      struct command_rx_struct *command, ...)
{

	xmlnode x;
	jpacket jp;
    at_buddy buddy;
	at_session s;
    ati ti;
    char *sn;
    struct aim_userinfo_s *userinfo;
    
    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);
            
	s = (at_session)ass->aux_data;
    ti = s->ti;
    sn = userinfo->sn;
    buddy = (at_buddy)xhash_get(s->buddies, sn);
    if(buddy == NULL)
    {
        jid jtmp;

        buddy = pmalloco(s->p, (sizeof(_at_buddy)));
        buddy->full = jid_new(s->p, ti->i->id);
        jid_set(buddy->full, sn, JID_USER);
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());

        xhash_put(s->buddies, buddy->full->user, buddy);
    }

    buddy->is_away = -1;

    xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());    

	x = xmlnode_new_tag("presence");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", ti->i->id);
	xmlnode_put_attrib(x, "type","unavailable");
	jp = jpacket_new(x);
	jid_set(jp->from, at_normalize(sn), JID_USER);
	xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));

    deliver(dpacket_new(jp->x), ti->i);

	return 1;
}

int at_parse_evilnotify(struct aim_session_t *sess, 
                              struct command_rx_struct *command, ...)
{
    va_list ap;
    xmlnode x;
    xmlnode body;
    jpacket jp;
    at_session s;
    ati ti;
    char msg[100];
    int newevil;
    struct aim_userinfo_s *userinfo;

    va_start(ap, command);
    newevil = va_arg(ap, int);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    memset(msg, '\0', 100);
    snprintf(msg, 100, "Warning from: %s (new level: %2.1f%%", 
             (userinfo && strlen(userinfo->sn))?userinfo->sn:"anonymous",
             ((float)newevil)/10);

    s = (at_session)sess->aux_data;
    ti = s->ti;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", jid_full(s->from));
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");
    xmlnode_insert_cdata(body, (char *)&msg, strlen((char *)&msg));

    jp = jpacket_new(x);
    deliver(dpacket_new(jp->x), ti->i);

    return 1;
}

int at_parse_userinfo(struct aim_session_t *sess, 
                            struct command_rx_struct *command, ...)
{
    struct aim_userinfo_s *userinfo;
    char *prof_encoding = NULL;
    char *prof = NULL;
    unsigned short inforeq = 0;
    xmlnode x;
    xmlnode show;
    xmlnode status;
    jpacket jp;
    at_session s;
    ati ti;

    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    prof_encoding = va_arg(ap, char *);
    prof = va_arg(ap, char *);
    inforeq = va_arg(ap, int);
    va_end(ap);

    s = (at_session)sess->aux_data;
    ti = s->ti;

    if (inforeq == AIM_GETINFO_GENERALINFO) {
        /*
         printf("faimtest: userinfo: profile_encoding: %s\n", prof_encoding ? prof_encoding : "[none]");
        printf("faimtest: userinfo: prof: %s\n", prof ? prof : "[none]");
        */
    } else if (inforeq == AIM_GETINFO_AWAYMESSAGE) {
        x = xmlnode_new_tag("presence");
	    xmlnode_put_attrib(x, "to", jid_full(s->cur));
    	xmlnode_put_attrib(x, "from", ti->i->id);
	    jp = jpacket_new(x);
    	jid_set(jp->from, at_normalize(userinfo->sn), JID_USER);
	    xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));
        show = xmlnode_insert_tag(x, "show");
        xmlnode_insert_cdata(show, "away", -1);
        status = xmlnode_insert_tag(x, "status");
        if(prof != NULL)
        {
            prof = strip_html(prof, jp->p);
            prof = str_to_UTF8(jp->p, prof);
            xmlnode_insert_cdata(status, prof, -1);
        } else {
            xmlnode_insert_cdata(status, "Away", -1);
        }
        deliver(dpacket_new(jp->x), ti->i);
    } else 
        log_debug(ZONE, "[AT] userinfo: unknown info request");

    return 1;
}


int at_parse_locerr(struct aim_session_t *sess, 
                          struct command_rx_struct *command, ...)
{
/* XXX Do we want this?
    va_list ap;
    char *destsn;
    unsigned short reason;
    char *sn;
    xmlnode x;
    xmlnode body;
    jpacket jp;
    session s;
    char msg[1024];

    va_start(ap, command);
    reason = va_arg(ap, unsigned short);
    destsn = va_arg(ap, char *);
    va_end(ap);

    memset(&msg, '\0', 1024);
    snprintf("User information for %s unavailable (reason 0x%04x: %s)\n", destsn, reason, (reason<msgerrreasonslen)?msgerrreasons[reason]:"unknown");

    s = (session)ass->aux_data;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->j));
	xmlnode_put_attrib(x, "from", aimtrans__instance->id);
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");
    xmlnode_insert_cdata(body, &msg, strlen(&msg));

    jp = jpacket_new(x);
	ehandler_send(ehandler_at, jp->x, s->j->server);
*/  
    return 1;
}

==> aim-transport-0.9.21.tar.gz: aim-transport-0.9.21/src/buddies.c <==
#include "aimtrans.h"

struct buddy_clean_data
{
    ati ti;
    xmlnode node;
};

void _at_buddies_unsubscribe(xht ht, const char *key, void *data, void * arg)
{
    xmlnode pres;
    at_session s;
    ati ti;

    s = (at_session)arg;
    ti = s->ti;

    pres = jutil_presnew(JPACKET__UNSUBSCRIBE, jid_full(s->cur), "Transport Removal");
    xmlnode_put_attrib(pres, "from", key);
    deliver(dpacket_new(pres), ti->i);
}

result at_buddy_pending_clean(void *arg)
{
    struct buddy_clean_data *bcd;
    ati ti;
    char *user;
    xmlnode node;

    bcd = (struct buddy_clean_data *)arg;
    node = bcd->node;
    ti = bcd->ti;
    user = xmlnode_get_attrib(node, "user");

    log_debug(ZONE, "[AT] Cleaning pending for %s: %s", user, xmlnode2str(node));
    pth_mutex_acquire(&ti->buddies_mutex, 0, NULL);

    xhash_zap(ti->pending__buddies, user);

    xmlnode_free(node);
    pth_mutex_release(&ti->buddies_mutex);

    return r_UNREG;
}

void at_buddy_addtolist(at_session s, spool sp, xmlnode x)
{
    at_buddy new;
    at_buddy old;
    xmlnode item;

    for(item = xmlnode_get_firstchild(x); item != NULL; 
                    item = xmlnode_get_nextsibling(item))
    {
        char *sn;

        sn = at_normalize(xmlnode_get_attrib(item, "name"));
        old = xhash_get(s->buddies, sn);
        if(old != NULL)
        {
            log_debug(ZONE, "[AT] We already have %s in our list", sn);
            continue;
        }

 		log_debug(ZONE, "[AIM] Adding a buddy %s from xdb", sn);

        spooler(sp, sn, "&", sp);
        new = pmalloco(s->p, sizeof(_at_buddy));
        new->full = jid_new(s->p, s->ti->i->id);
        jid_set(new->full, sn, JID_USER);
        new->last = xmlnode_new_tag_pool(s->p,"query");
        new->is_away = -1;
        xmlnode_put_attrib(new->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(new->last,"stamp",jutil_timestamp()); 

        xhash_put(s->buddies, new->full->user, new);
    }
}

char *at_buddy_buildlist(at_session s, jid from)
{
    char *list;
	char *blist;
	spool sp;
    pool p, bp;
    xmlnode x;
    xmlnode msg;

    p = pool_new();
	sp = spool_new(p);

    x = at_xdb_get(s->ti, from, AT_NS_ROSTER);
    if(x != NULL)
    {
        at_buddy_addtolist(s, sp, x);
    }

    x = xhash_get(s->ti->pending__buddies, jid_full(from));
    if(x == NULL)
    {
        return NULL;
    } else {
        at_buddy_addtolist(s, sp, x);
    }
    
    list = spool_print(sp);
    if(list)
    {
        blist = strdup(list);
    } else {
        blist = NULL;
    }
    log_debug(ZONE, "[AT] Buddylist generation complete");

    pool_free(p);

    return blist;
}

int at_buddy_subscribe(ati ti, jpacket jp)
{
    xmlnode dup,dup2;
    at_session s=at_session_find_by_jid(ti, jp->from);

    /* check for a valid session */
    if(s==NULL)
    {
        xmlnode err,error;
        /* if no session, and subscribing to a buddy, not aim-t */
        err=xmlnode_new_tag("message");
        xmlnode_put_attrib(err,"type","error");
        xmlnode_put_attrib(err,"from",ti->i->id);
        xmlnode_put_attrib(err,"to",jid_full(jp->from));
        error=xmlnode_insert_tag(err,"error");
        xmlnode_insert_cdata(error,"Cannot Subscribe to a AIM Buddy without a registration",-1);
        xmlnode_put_attrib(error,"code","407");
        deliver(dpacket_new(err), ti->i);

        return 0;
    }
    /* XXX
	 * Create the duplicate for the subscribe notice
	 * Construct a subscribed notification
	 * 
	 * I had to do these in this odd order due to memory concerns
	 * Dunno why I didn't just make two dups.
	 *
	 */
	dup = xmlnode_dup(jp->x);
	dup2 = xmlnode_dup(jp->x);

	xmlnode_put_attrib(dup, "to", jid_full(jp->from));
	xmlnode_put_attrib(dup, "from", jid_full(jp->to));
	xmlnode_put_attrib(dup, "type", "subscribed");
	log_debug(ZONE, "[AIM] Sending subscribed notice\n");
    deliver(dpacket_new(dup), ti->i);

	/* Construct the subscribe back */
	xmlnode_put_attrib(dup2, "type", "subscribe");
    xmlnode_put_attrib(dup2, "to", jid_full(jp->from));
    xmlnode_put_attrib(dup2, "from", jid_full(jp->to));
	log_debug(ZONE, "[AIM] Asking for a subscribe\n");
    deliver(dpacket_new(dup2), ti->i);

	return 0;
}

int at_buddy_add(ati ti, jpacket jp)
{
	at_session s;
	at_buddy buddy;
    pool p;
	int newbud=0;
    struct buddy_clean_data *bcd;
	
    log_debug(ZONE, "[AIM] Going to add a buddy\n");
	s = at_session_find_by_jid(ti, jp->from);
	if(s==NULL || s->loggedin == 0) 
    {
        xmlnode cur, item;
        
        pth_mutex_acquire(&ti->buddies_mutex, 0, NULL);
        cur = xhash_get(ti->pending__buddies, jid_full(jp->from));
        if(cur == NULL)
        {
            cur = xmlnode_new_tag("buddies");
            xmlnode_put_attrib(cur, "user", jid_full(jp->from));
            bcd = pmalloco(xmlnode_pool(cur), sizeof(struct buddy_clean_data));
            bcd->node = cur;
            bcd->ti = ti;
            register_beat(30, at_buddy_pending_clean, bcd);
        }
        item = xmlnode_insert_tag(cur, "item");
        xmlnode_put_attrib(item, "name", jp->to->user);
        xmlnode_free(jp->x);
        log_debug(ZONE, "[AT] Pending for %s: %s", xmlnode_get_attrib(cur, "user"), xmlnode2str(cur));
        xhash_put(ti->pending__buddies, xmlnode_get_attrib(cur, "user"), cur);
        pth_mutex_release(&ti->buddies_mutex);
        return 1;
	}
    
    if(xhash_get(s->buddies, jp->to->user) == NULL)
    {
		log_debug(ZONE, "[AIM] Adding buddy %s to %s\n", jp->to->user, jid_full(jp->from));
        buddy = pmalloco(s->p, sizeof(_at_buddy));
        buddy->full=jid_new(s->p, jid_full(jp->to));
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        buddy->is_away = -1;
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());
        xhash_put(s->buddies, buddy->full->user, buddy);
        at_buddy_subscribe(ti, jp);
    }
    aim_add_buddy(s->ass, aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS),
                  jp->to->user);
    
    xmlnode_free(jp->x);
    return 1;
}

int at_parse_oncoming(struct aim_session_t *ass, 
                      struct command_rx_struct *command, ...)
{

    xmlnode away;
    xmlnode status;
	xmlnode x;
	at_session s;
    at_buddy buddy;
    ati ti;
    jpacket jp;
    char *msg;
    struct aim_userinfo_s *userinfo;
    int was_away;

    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    log_debug(ZONE, "Oncoming buddy %s", userinfo->sn);

	s = (at_session)ass->aux_data;
    ti = s->ti;

    buddy = xhash_get(s->buddies, userinfo->sn);
    if(buddy == NULL)
    {
        jid jtmp;

        buddy = pmalloco(s->p, (sizeof(_at_buddy)));
        buddy->full = jid_new(s->p, ti->i->id);
        jid_set(buddy->full, userinfo->sn, JID_USER);
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        buddy->is_away = -1;
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());

        xhash_put(s->buddies, buddy->full->user, buddy);
    }
    if(buddy->login_time == 0)
    {
        buddy->login_time = userinfo->onlinesince;
    }
    buddy->idle_time = userinfo->idletime;

    was_away = buddy->is_away;
    if(userinfo->flags&AIM_FLAG_AWAY) 
        buddy->is_away = 1;
    else
        buddy->is_away = 0;

    if(buddy->is_away == 0 && (buddy->is_away != was_away))
    {
        char *status_msg;

        x = xmlnode_new_tag("presence");
	    xmlnode_put_attrib(x, "to", jid_full(s->cur));
    	xmlnode_put_attrib(x, "from", ti->i->id);
	    jp = jpacket_new(x);
    	jid_set(jp->from, at_normalize(userinfo->sn), JID_USER);
	    xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));
        status = xmlnode_insert_tag(x, "status");
        status_msg = pmalloco(xmlnode_pool(x), 30);
        sprintf(status_msg, "Online (Idle %d Seconds)", buddy->idle_time);
        xmlnode_insert_cdata(status, status_msg, -1);
        deliver(dpacket_new(jp->x), ti->i);
    } 
    else if(buddy->is_away == 1 && (buddy->is_away != was_away))
    {
        log_debug(ZONE, "[AT] Requesting Away message for %s", userinfo->sn);
        aim_getinfo(ass, command->conn, userinfo->sn, AIM_GETINFO_AWAYMESSAGE);
    }

	return 1;
}

int at_parse_offgoing(struct aim_session_t *ass, 
                      struct command_rx_struct *command, ...)
{

	xmlnode x;
	jpacket jp;
    at_buddy buddy;
	at_session s;
    ati ti;
    char *sn;
    struct aim_userinfo_s *userinfo;
    
    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);
            
	s = (at_session)ass->aux_data;
    ti = s->ti;
    sn = userinfo->sn;
    buddy = (at_buddy)xhash_get(s->buddies, sn);
    if(buddy == NULL)
    {
        jid jtmp;

        buddy = pmalloco(s->p, (sizeof(_at_buddy)));
        buddy->full = jid_new(s->p, ti->i->id);
        jid_set(buddy->full, sn, JID_USER);
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());

        xhash_put(s->buddies, buddy->full->user, buddy);
    }

    buddy->is_away = -1;

    xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());    

	x = xmlnode_new_tag("presence");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", ti->i->id);
	xmlnode_put_attrib(x, "type","unavailable");
	jp = jpacket_new(x);
	jid_set(jp->from, at_normalize(sn), JID_USER);
	xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));

    deliver(dpacket_new(jp->x), ti->i);

	return 1;
}

int at_parse_evilnotify(struct aim_session_t *sess, 
                              struct command_rx_struct *command, ...)
{
    va_list ap;
    xmlnode x;
    xmlnode body;
    jpacket jp;
    at_session s;
    ati ti;
    char msg[100];
    int newevil;
    struct aim_userinfo_s *userinfo;

    va_start(ap, command);
    newevil = va_arg(ap, int);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    memset(msg, '\0', 100);
    snprintf(msg, 100, "Warning from: %s (new level: %2.1f%%", 
             (userinfo && strlen(userinfo->sn))?userinfo->sn:"anonymous",
             ((float)newevil)/10);

    s = (at_session)sess->aux_data;
    ti = s->ti;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", jid_full(s->from));
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");
    xmlnode_insert_cdata(body, (char *)&msg, strlen((char *)&msg));

    jp = jpacket_new(x);
    deliver(dpacket_new(jp->x), ti->i);

    return 1;
}

int at_parse_userinfo(struct aim_session_t *sess, 
                            struct command_rx_struct *command, ...)
{
    struct aim_userinfo_s *userinfo;
    char *prof_encoding = NULL;
    char *prof = NULL;
    unsigned short inforeq = 0;
    xmlnode x;
    xmlnode show;
    xmlnode status;
    jpacket jp;
    at_session s;
    ati ti;

    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    prof_encoding = va_arg(ap, char *);
    prof = va_arg(ap, char *);
    inforeq = va_arg(ap, int);
    va_end(ap);

    s = (at_session)sess->aux_data;
    ti = s->ti;

    if (inforeq == AIM_GETINFO_GENERALINFO) {
        /*
         printf("faimtest: userinfo: profile_encoding: %s\n", prof_encoding ? prof_encoding : "[none]");
        printf("faimtest: userinfo: prof: %s\n", prof ? prof : "[none]");
        */
    } else if (inforeq == AIM_GETINFO_AWAYMESSAGE) {
        x = xmlnode_new_tag("presence");
	    xmlnode_put_attrib(x, "to", jid_full(s->cur));
    	xmlnode_put_attrib(x, "from", ti->i->id);
	    jp = jpacket_new(x);
    	jid_set(jp->from, at_normalize(userinfo->sn), JID_USER);
	    xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));
        show = xmlnode_insert_tag(x, "show");
        xmlnode_insert_cdata(show, "away", -1);
        status = xmlnode_insert_tag(x, "status");
        if(prof != NULL)
        {
            prof = strip_html(prof, jp->p);
            prof = str_to_UTF8(jp->p, prof);
            xmlnode_insert_cdata(status, prof, -1);
        } else {
            xmlnode_insert_cdata(status, "Away", -1);
        }
        deliver(dpacket_new(jp->x), ti->i);
    } else 
        log_debug(ZONE, "[AT] userinfo: unknown info request");

    return 1;
}


int at_parse_locerr(struct aim_session_t *sess, 
                          struct command_rx_struct *command, ...)
{
/* XXX Do we want this?
    va_list ap;
    char *destsn;
    unsigned short reason;
    char *sn;
    xmlnode x;
    xmlnode body;
    jpacket jp;
    session s;
    char msg[1024];

    va_start(ap, command);
    reason = va_arg(ap, unsigned short);
    destsn = va_arg(ap, char *);
    va_end(ap);

    memset(&msg, '\0', 1024);
    snprintf("User information for %s unavailable (reason 0x%04x: %s)\n", destsn, reason, (reason<msgerrreasonslen)?msgerrreasons[reason]:"unknown");

    s = (session)ass->aux_data;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->j));
	xmlnode_put_attrib(x, "from", aimtrans__instance->id);
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");
    xmlnode_insert_cdata(body, &msg, strlen(&msg));

    jp = jpacket_new(x);
	ehandler_send(ehandler_at, jp->x, s->j->server);
*/  
    return 1;
}
